(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/Absyn.sml,v
   Comment: High level abstract syntax based on wasowski/sestoft:2002.
   Release: scope-0-16_final

     Copyright (C) 2001-2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


   This is highlevel abstract syntax which is quite far away from concrete
   textual vsr syntax. It can be understood as IR suitable for interpretation
   or code generation, or perhaps abstract syntax of concrete visual syntax.
   It is much more conceptaul than textual description of statechart.

   CHANGES (comparing to VS):
   
   Grammar of exp, sexp and oexp and accesses has been generalized
   (orthogonalized). Arbitrary expressions may be now used as parameters,
   indexers etc.

   PROBLEMS:

   1. Variable ranges are simplified (not used currently)
   2. No Do activities
   3. No internal transitions (this should be resolved before)
   
   Absyn.sml,v 1.95 2004/05/13 15:26:30 wasowski Exp
*)

structure Absyn = struct

open Common Err Haskellish Say
infixr 0 $

  
exception StateNotFound;
exception StateTypeMismatch;

(* states and hierarchy *)
 
datatype marking = History | Initial 

datatype 'aor orstate = ORState of { 
			name       : stateid,			      
			parent     : stateid,
			children   :'aor andstate list,
			marking    : marking,	  (* lifted up one level *)			      
			annotation :'aor
		  }						      
		  
and 'aor andstate = ANDState of {
			name     : stateid,
			parent   : stateid option,      (* NONE for root *)
			children :'aor orstate list,
			entry    : action,
			exit     : action
		  }

  
(* Union of state types. double constructors are unfortunately unavoidable
   because record types cannot be recursive themselves in SML97 *)

datatype 'aor state = InclOR  of 'aor orstate
                    | InclAND of 'aor andstate

(* and to simple helpers to make constructing more succint *)
	 
fun inor  st = InclOR  (ORState st)
fun inand st = InclAND (ANDState st)

(* In addition to state hierarchy I maintain state environment which is a
   flat index into hierarchy. This relies on the fact that hierarchy is
   persistent - static during execution. Similarly references to parent
   states could be embedded in the hierarchy in C style to allow movements
   up and down. *)

type 'aor stenv = 'aor state env

(* hierarchy is modelled as the proper hierarchy starting with root state
   and an index to it *)

type 'aor hierarchy = 'aor andstate * 'aor stenv
 
(* The index was not explicitly modelled in theoretical description of TR since
   it is only introduced to improve efficiency. The index is built by a
   function idx_hierarchy below. *)

fun getstate (h :'aor hierarchy) (s :stateid) :'aor state = SE.lookup (#2 h) s
  
(* Combinators on functions acceping states --------------------------------- *)

fun lift_orand (fo :'aor orstate -> 'a) (fa :'aor andstate -> 'a) 
					:'aor state -> 'a =
      (fn s => case s of InclOR  so => fo so | InclAND sa => fa sa)
      
fun lift_h (f :'aor state -> 'a) : 'aor hierarchy -> stateid -> 'a =
      fn h => fn sid => f (getstate h sid) 
      
fun lift_st (f : 'a -> 'b ) (g : 'a * 'b ->'b) = fn a => g (a, f a) 

fun id (x :'a) :'a = x




(* getters/setters for hierarchy -------------------------------------------- *)
  
(* names of states *)
  
fun nameor  (ORState  s :'aor orstate)  :stateid = #name s
fun nameand (ANDState s :'aor andstate) :stateid = #name s
fun name (s :'aor state) :stateid = lift_orand nameor nameand s

fun setannotationeor (ORState s:'aor orstate) (a :'aor):'aor orstate =
	ORState { name= #name s, parent= #parent s, children= #children s,
		  marking= #marking s, annotation= a }
fun annotationor (ORState {annotation,...}:'aor orstate) :'aor =
	annotation


(* entry/exit actions *)

fun exand (ANDState s :'aor andstate) = #exit  s
fun enand (ANDState s :'aor andstate) = #entry s

(* children *)

fun childrenand (ANDState s) :'aor orstate list  = #children s
fun childrenor  (ORState s)  :'aor andstate list = #children s
fun children (s: 'aor state) :'aor state list = 
	lift_orand (map InclAND o childrenor) (map InclOR o childrenand) s
fun childrenh (h: 'aor hierarchy) : stateid -> 'aor state list =
	lift_h children h

fun setchildrenand (ANDState s :'aor1 andstate) (kids :'aor2 orstate list) 
							:'aor2 andstate =
	ANDState { name= #name s, parent= #parent s, children= kids,
		   entry= #entry s, exit= #exit s }

(* parent state *)

fun parentor  (ORState s)  :stateid = #parent s
fun parentand (ANDState s) :stateid = 
      case #parent s of SOME p => p | NONE => raise StateNotFound
fun parent  s = lift_orand parentor parentand s
fun parenth h = lift_h parent h
  


(* higher order functions on hierarchy -------------------------------------- *)

(* map allows you to transform states in hierarchy (but not the shape)        *)
(* fo, fa should not modify children of the state unless you want to go mad   *)

fun hmapor (fo :'aor1 orstate -> 'aor2 orstate) 
           (fa :'aor1 andstate -> 'aor2 andstate) 
	   (s :'aor1 orstate) :'aor2 orstate = 
	let val ORState s1 = fo s
	in ORState { 
		name= #name s1, parent= #parent s1, 
		children= map (hmapand fo fa) (childrenor s),
		marking= #marking s1, annotation= #annotation s1}
	end

and hmapand (fo :'aor1 orstate -> 'aor2 orstate) 
            (fa :'aor1 andstate -> 'aor2 andstate) 
	    (s :'aor1 andstate) :'aor2 andstate =
	let val ANDState s1 = fa s
	in ANDState { name= #name s1, parent= #parent s1, 
		      children= map (hmapor fo fa) (childrenand s),
		      entry= #entry s1, exit= #exit s1 }
	end

(* This is a simpler version which only maps the annotations *)

fun amapor (fo :'aor1 -> 'aor2) (ORState s :'aor1 orstate) :'aor2 orstate =
	ORState { name= #name s, parent= #parent s, 
		  children= map (amapand fo) (#children s),
		  marking= #marking s, annotation= fo (#annotation s) }

and amapand (fo :'aor1 -> 'aor2) (ANDState s:'aor1 andstate) :'aor2 andstate =
	ANDState { name= #name s, parent= #parent s,
		   children= map (amapor fo) (#children s),
		   exit= #exit s, entry= #entry s }


(* general fold over hierarchy.  This fold has a starting point e
   used for children of basic states. So actually fa(s,e) is called if s is a
   basic state.

   This is generalized view on folds over hierarchy tree.  Fold function
   applies fa and fo function on the hierarchy in X-order and Y direction,
   where X stands for pre or post determining if node itself is traversed
   before or after its children and Y stands for R->L or L->R and stands
   for direction of iterating children of the node.  
   
   Only some combinations of X and Y are useful currently. I have
   implemented specialized versions explicitly. Both for efficiency and
   expressivness. Using hofs without module languege seems to lead to ugly
   solution. One could use functors ut this would break current state of
   this module (structure) or forced me to factor folds out which I want to
   avoid now. *)

type ('a,'aor) oporty = 'aor orstate  * 'a -> 'a
type ('a,'aor) opandty = 'aor andstate * 'a -> 'a

type ('a,'aor) foldorty = 
	('a,'aor) oporty -> ('a,'aor) opandty -> 'a -> 'aor orstate -> 'a
type ('a,'aor) foldandty = 
	('a,'aor) oporty -> ('a,'aor) opandty -> 'a -> 'aor andstate -> 'a
type ('a,'aor) foldty = 
	('a,'aor) oporty -> ('a,'aor) opandty -> 'a -> 'aor state -> 'a

(*type ('st,'e,'aor) foldty = ('st * 'e -> 'e) -> 'e -> 'st list -> 'e *)
  
(* lifts type of fold from or/and-state to state *)

fun hfold_lift (foldor :('a,'aor) foldorty) (foldand :('a,'aor) foldandty)
	(fo :('a,'aor) oporty) (fa: ('a,'aor) opandty) (e :'a) (s :'aor state) :'a =
      case s of InclOR  s => foldor fo fa e s
	| InclAND s => foldand fo fa e s

(* mixed version: post order, R -> L *)
  
fun hfoldor (fo :('a,'aor) oporty) (fa :('a,'aor) opandty) (e : 'a) (s :'aor orstate) : 'a = 
     fo (s,foldr (fn (s',e')=> hfoldand fo fa e' s') e (childrenor s))

and hfoldand (fo :('a,'aor) oporty) (fa :('a,'aor) opandty) (e : 'a) (s :'aor andstate) : 'a = 
     fa (s,foldr (fn (s',e')=> hfoldor fo fa e' s') e (childrenand s)) 

fun hfold fo = hfold_lift hfoldor hfoldand fo


(* R->L preorder *)

fun hfoldror (fo :('a,'aor) oporty) (fa :('a,'aor) opandty) (e : 'a) (s :'aor orstate) : 'a = 
     foldr (fn (s',e')=> hfoldrand fo fa e' s') (fo (s,e)) (childrenor s )

and hfoldrand (fo :('a,'aor) oporty) (fa :('a,'aor) opandty) (e : 'a) (s :'aor andstate) : 'a = 
     foldr (fn (s',e')=> hfoldror fo fa e' s' ) (fa (s,e)) (childrenand s)

fun hfoldr fo = hfold_lift hfoldror hfoldrand fo


(* L->R postorder *)

fun hfoldlor (fo :('a,'aor) oporty) (fa :('a,'aor) opandty) (e : 'a)
					       (s :'aor orstate) : 'a = 
       fo (s,foldl (fn (s',e')=> hfoldland fo fa e' s') e (childrenor s))

and hfoldland (fo :('a,'aor) oporty) (fa :('a,'aor) opandty) (e : 'a)
						   (s :'aor andstate) : 'a = 
       fa (s,foldl (fn (s',e')=> hfoldlor  fo fa e' s') e (childrenand s))

fun hfoldl fo = hfold_lift hfoldlor hfoldland fo











(* Comparison on states: mainly used to build efficient container types *)

fun compareor (s1 :'aor orstate, s2 :'aor orstate) :order = 
	Symbol.compare (nameor s1,nameor s2)
	
fun compareand(s1 :'aor andstate, s2 :'aor andstate) :order = 
	Symbol.compare (nameand s1,nameand s2)

fun compare   (s1 :'aor state, s2 :'aor state) :order =
	Symbol.compare (name s1, name s2)

(* get names of all and all basic states in hierarchy *)

fun all_names( (_,env) :'aor hierarchy ) : symbol list = 
    map #1 (SE.toList env)

fun basic_names( (s,_) :'aor hierarchy ) : symbol list = 
    let fun fa (ANDState {children=[],name,...}, e) = name::e
          | fa (_,e) = e
    in hfoldland (fn (s,e)=>e) fa [] s end
    
    

  
(* transitions -------------------------------------------------------------- *)

(* 'atr - whole transition's annotation            *)

type 'atr tran = { 
	
	trigger    : trigger,
	scope      : stateid,				(* always orstate   *)
	pos        : stateid list,			(* always andstates *)
	neg        : stateid list,			(* always andstates *)
	guard      : guard,
	action     : action,
	targets    : stateid list,			(* always andstates *)
	annotation :'atr
}

(* annotation type for scoped transitions (after static scope) *)

(* FIXME: also guard annotation should be somehow included in this, so
   BDDs are not reconstructed repeatedly *)

type static = { stsrcs  :stateid list option,
		stscope :stateid,
                tgts :stateid list } list 

		(* targets may not be maximally static, which easily obtained
		   during flattening, but nondesirable for hierarchical code
		   generation *)

(* some setters/getters for trans *)

fun setaction (t :'atr tran) a :'atr tran = 
	{ trigger= #trigger t, scope= #scope t, pos= #pos t, neg= #neg t,
	  guard= #guard t, action= a, targets= #targets t, 
	  annotation= #annotation t}

fun setannotationtr (t :'atr1 tran) (a :'atr2) :'atr2 tran = 
	{ trigger= #trigger t, scope= #scope t, pos= #pos t, neg= #neg t,
	  guard= #guard t, action= #action t, targets= #targets t,
	  annotation= a }
		 
(* model *)
  

type ('aor,'atr) system = {

	name      : symbol,
	fname     : symbol,		(* name of the source file          *)
	hierarchy : 'aor hierarchy,	(* the whole state business is here *)
	vartypes  : vartype  env,
	evtypes   : evtype   env,
	evgroups  : evgroup  env,
	otypes    : otype    env,
	denv      : denv,		(* mapping of default states        *)
	venv      : venv,		(* initial variable assignment      *)

	cenv      : const env,		(* named constants                  *)
	trans     : 'atr tran list
  }

(* type of complete models *)

type ('aor,'atr) model = ('aor,'atr) system list

(* some getters/setters for system *)

(* get external variables *)

fun get_extv (sys :('aor,'atr) system) :var list = 
	let val vt = SE.toList (#vartypes sys)
	    val vt1= List.filter (fn (v,(i,t))=> not i) vt
	in map #1 vt1 end


fun setevtypes (sys : ('aor,'atr) system) evtypes :('aor,'atr) system = {
	name= #name sys, fname= #fname sys, hierarchy= #hierarchy sys,
	vartypes= #vartypes sys, evtypes= evtypes, evgroups= #evgroups sys,
	otypes= #otypes sys, denv = #denv sys,venv= #venv sys,cenv= #cenv sys, 
	trans= #trans sys
  }

fun setotypes (sys :('aor,'atr) system) otypes :('aor,'atr) system = {
	name= #name sys, fname= #fname sys, hierarchy= #hierarchy sys,
	vartypes= #vartypes sys, evtypes= #evtypes sys, evgroups= #evgroups sys,
	otypes= otypes, denv = #denv sys,venv= #venv sys,cenv= #cenv sys, 
	trans= #trans sys
  }

fun settrans (sys :('aor,'atr1) system) (trans :'atr2 tran list) :('aor,'atr2) system = {
	name= #name sys, fname= #fname sys, hierarchy= #hierarchy sys,
	vartypes= #vartypes sys, evtypes= #evtypes sys, evgroups= #evgroups sys,
	otypes= #otypes sys, denv = #denv sys,venv= #venv sys,cenv= #cenv sys, 
	trans= trans
  }

fun setevgroups (sys :('aor,'atr) system) evgroups :('aor,'atr) system = {
	name= #name sys, fname= #fname sys, hierarchy= #hierarchy sys,
	vartypes= #vartypes sys, evtypes= #evtypes sys, evgroups= evgroups,
	otypes= #otypes sys, denv = #denv sys,venv= #venv sys,cenv= #cenv sys, 
	trans= #trans sys 
  }

fun sethierarchy (sys :('aor1,'atr) system) (h :'aor2 hierarchy) 
						:('aor2,'atr) system = 
  {
	name= #name sys, fname= #fname sys, hierarchy= h,
	vartypes= #vartypes sys, evtypes= #evtypes sys, evgroups= #evgroups sys,
	otypes= #otypes sys, denv = #denv sys,venv= #venv sys,cenv= #cenv sys, 
	trans= #trans sys 
  }

(* -------------------------------------------------------------------------- *)
(* All types are done until now. Auxiliary functions follow                   *)

fun getand (InclAND s) = SOME s | getand _ = NONE
fun getor  (InclOR  s) = SOME s | getor  _ = NONE

fun getandh (h :'aor hierarchy) (s :stateid) = valOf (getand (getstate h s))
fun getorh  (h :'aor hierarchy) (s :stateid) = valOf (getor  (getstate h s))


fun gettree (h :'aor hierarchy) :'aor andstate = #1 h
fun getroot (h :'aor hierarchy) :'aor state = InclAND (gettree h)

(* siblings - including self *)

fun siblingsst (h :'aor hierarchy) (s :'aor state) :'aor state list =
	children $ getstate h $ parent s

fun siblingandst (h :'aor hierarchy) (s :'aor andstate) :'aor andstate list =
	childrenor $ valOf $ getor $ getstate h $ parentand s
	handle Option.Option => impossible "should always be orstate"

fun isrootand (ANDState s) = #parent s = NONE		(* FIXME wywal *)
fun isroot s  = lift_orand (fn _ => false) isrootand s	(* FIXME wywal *)		
fun isrooth h = lift_h isroot h				(* FIXME wywal *)

fun ishistoryor (ORState s) = #marking s = History
fun ishistory s  = lift_orand ishistoryor (fn _ => raise StateTypeMismatch) s
fun ishistoryh h = lift_h ishistory h

(* state types *)

fun isand   (InclAND s) = true | isand _ = false
fun isor    (InclOR  s) = true | isor  _ = false
  
fun isandh h = lift_h isand h
fun isorh  h = lift_h isor h
  
  
fun isbasicand (ANDState s) = null (#children s)
fun isbasic s = lift_orand (fn _ => false) isbasicand s

 
    
(* hierarchy constructor -- buids an index to explicit hierarchy starting
   with root *)  

fun idx_hierarchy (root : 'aor andstate) : 'aor hierarchy = 
      let
	fun idxor stenv (ORState s) : 'aor stenv = 
		let 
			val stenv1 = SE.bind1 stenv (#name s, inor s)
		in 
			foldl (fn (a,e) => idxand e a) 
				stenv1 (#children s) 
		end
	and idxand stenv (ANDState s) : 'aor stenv =
		let 
			val stenv1 = SE.bind1 stenv (#name s, inand s)
		in	
			foldl (fn (a,e) => idxor e a) 
				stenv1 (#children s)
		end
      in (root, idxand empty root) end handle e => intErr e 200305074

(* rebuild default environment after hierarchy was rebuilt. Unfortunately
   denv uses persistent references to andstates in hierarchy (instead of
   names) which makes model tranformations troublesome. This function takes
   an old denv and new (incompatible) hierarchy and builds new denv. It
   relies on the fact that state names are unique across the hierarchy! *)

fun mkdenv (denv :denv) (root :'aor andstate) :denv =
      let fun mkand (s :'aor andstate, ndenv :denv) = 
		foldl mkor ndenv (childrenand s)
          and mkor  (s :'aor orstate,  ndenv :denv) =
		let val dname =	SE.lookup denv (nameor s)
		    val dstop = List.find (fn s => dname = nameand s) 
					(childrenor s)
		    val dst   = valOf dstop handle Option.Option => 
				impossible "#200305073"
		    val ndenv1= SE.bind1 ndenv (nameor s,nameand dst) 
		in foldl mkand ndenv1 (childrenor s) end
      in mkand (root, empty) end

(* more properties of hierarchy *)

(* list of ancestors is sorted with age decreasing -- root is the prefix *)	

local 
      fun recancest h s ancs :stateid list =
	let val p = parenth h s 
	in recancest h p (p::ancs) end handle StateNotFound => ancs
in
      fun ancesth (h :'aor hierarchy) (s :stateid) :stateid list = 
	recancest h s []
	
      fun ancesthst (h :'aor hierarchy) (s :stateid) :stateid list =    
	recancest h s [s]
end

(* same as above, but only return and-states if applied to an andstate
   and only or-states if applied to an or-state. *)

local 
      fun recancest h s ancs :stateid list =
	let val p = parenth h (parenth h s)
	in recancest h p (p::ancs) end handle StateNotFound => ancs
in
      fun ancest2h (h :'aor hierarchy) (s :stateid) :stateid list = 
	recancest h s []
	
      fun ancest2hst (h :'aor hierarchy) (s :stateid) :stateid list =    
	recancest h s [s]
end

(* checks if s1 is an ancestor of s2.  FIXME: could be more efficient
   traversing hierarchy and checking ancestors instead of computing anestors
   set in place *)

fun ancestorh (h :'aor hierarchy) (s1 :stateid) (s2 :stateid) :bool =
	let val ancests = ancesth h s2
	in List.exists (fn a => a=s1) ancests end

fun ancestorhst (h :'aor hierarchy) (s1 :stateid) (s2 :stateid) :bool =
	s1 = s2 orelse ancestorh h s1 s2
	handle e =>intErr e 200307115



(* returns true if s1 and s2 are ancestrally related *)

fun ancestralh (h :'aor hierarchy) (s1 :stateid) (s2 :stateid) :bool =
	ancestorh h s1 s2 orelse ancestorh h s2 s1

fun ancestralhst (h :'aor hierarchy) (s1 :stateid) (s2 :stateid) :bool =
	s1 = s2 orelse ancestorh h s1 s2 orelse ancestorh h s2 s1
  

 
(* Lowest Common Ancestor *) 

(* Stephen and Christian have much better solution for the problem (which
   is said to be "optimal"). Note also that alternatively this can be
   defined for a pair of states and then folded. For practical reasons this
   may be a bit faster. Also we usually need it for a pair. *)

fun LCA (h :'aor hierarchy) [] :stateid = 
		impossible "LCA is not defined for empty set"
  | LCA (h :'aor hierarchy) (ss :stateid list) :stateid =
	let	
		val a = map (ancesthst h) ss 
		fun f (l :stateid list list) (lca :stateid) =
			let val hds = map hd l 
			    val fst = hd hds 
			in 
				if List.all (fn h => h=fst) hds 
				then f (map tl l) fst else lca
			end handle Empty => lca
	in f a (name $ getroot h) end handle e => intErr e 200312140
		

(* irreflexive version of LCA - it is guaranteed that result is a strong
   parent of all list elements. Undefined for any list containing topstate. *)

fun irrLCA (h :'aor hierarchy) [] :stateid = 
	impossible "irrLCA is not defined for empty set"
  | irrLCA (h :'aor hierarchy) (ss :stateid list) :stateid =
      let val a = map (ancesth h) ss 
          fun f (l :stateid list list) (lca :stateid) =
	let val hds = map hd l 
	    val fst = hd hds 
	in if List.all (fn h => h=fst) hds 
	   then f (map tl l) fst else lca
	end handle Empty => lca
      in f a (name $ getroot h) end handle e => intErr e 200307040

(* the most trivial and unefficient algorithm for testing orthogonality *)

fun orthogonal (h :'aor hierarchy) (s1 :stateid) (s2 :stateid) :bool =
	(not (ancestralh h s1 s2)) andalso isandh h (LCA h [s1,s2])

(* events set of transitions trigger *)	

fun events (eg :evgroup env) (t :'atr tran) :eventset =
	case #trigger t of
		  Event e	=> Set.singleton (#1 e)
		| EventGroup e	=> SE.lookup eg e

(* checks if e triggers t *)

fun triggers (eg :evgroup env) (e :event) (t :'atr tran) :bool =
	case #trigger t of
		  Event (e1,_) => e = e1 
		| EventGroup e1 => Set.member (SE.lookup eg e, e)

  
(* RightMost weak Descendant (RMD) of given orstate/andstate.           *)
(* The weakness means that for basic state it is the basic state itself *)

fun RMDor (s :'aor orstate) :'aor andstate = RMDand (List.last (childrenor s))

and RMDand (s :'aor andstate) :'aor andstate = if isbasicand s then s
			       else RMDor (List.last (childrenand s))

(* Left most descendant *)

fun LMDor  (s :'aor orstate ) :'aor andstate = LMDand (hd (childrenor s))
and LMDand (s :'aor andstate) :'aor andstate = if isbasicand s then s 
				       else LMDor (hd (childrenand s))

(* Crossed Descendant of given orstate/andstate.            *)
(* crossed bcs it selects the rightmost kid on orstates and *)
(* the leftmost kid on andstates                            *)
  

fun CDor (s :'aor orstate) :'aor andstate = CDand (hd (childrenor s))

and CDand (s :'aor andstate) :'aor andstate = if isbasicand s then s
				    else CDor (List.last (childrenand s))


  



(* sort transitions by priority. To achieve VS conflict resoltuion
   states are assigned priorities in preorder traverse and transitions are
   sorted by priorities of their scopes.  FIXME: Inefficient. For instance
   priorites could be explicitly kept in hierarchy (computed while it is
   being built) and then just sorted using Listsort. This inefficiency does
   not influence interpreting speed (and generated code). Function is only
   called once - during translation from vsr -> Absyn *)

fun prioritize (root :'aor andstate) (trans :'atr tran list) :'atr tran list =
      let
	fun fo (ORState s, tr) :'atr tran list = 
		List.filter (fn t => #scope t = (#name s)) trans @ tr
	fun fa (s,tr) :'atr tran list = tr
	val prioritized = hfoldand fo fa [] root 
	val _ = assert (fn () => length trans = length prioritized)
		"Some transitions has been lost during sorting."
      in prioritized end


(* Hierarchy tree annotation *)

type attr = { expure: bool,
              enpure: bool,
	      deeph : bool } 

fun expureor  (ORState s :attr orstate)   = #expure (#annotation s)
fun expureand (ANDState s :attr andstate) = List.all expureor (#children s)
fun expure s  = lift_orand expureor expureand s
fun expureh h = lift_h expure

fun enpureor  (ORState s  :attr orstate)  = #enpure (#annotation s)
fun enpureand (ANDState s :attr andstate) = List.all enpureor (#children s)
fun enpure s  = lift_orand enpureor enpureand s
fun enpureh h = lift_h enpure

fun deephor  (ORState s :attr orstate ) = #deeph (#annotation s)
fun deephand (ANDState s:attr andstate) = List.all deephor (#children s)
fun deeph s  = lift_orand deephor deephand s
fun deephh h = lift_h enpure
	      
(* deeph is set at *any* level of deep history (transitively) *)

(* FIXME: if annotations where 'aor ref type it would be much more 
   efficient (n reindexing). This would be better done on Absyn level. *)

fun annotate_or (ORState s :'aor orstate) : attr orstate =
	let val (kids,attrs) = ListPair.unzip (map annotate_and (#children s))
	    val attr = { expure = List.all #expure attrs, 
			 enpure = List.all #enpure attrs,
			 deeph  = #marking s = History andalso
				  List.all #deeph attrs }
	in ORState  { name= #name s, parent= #parent s, children= kids,
		      marking= #marking s, annotation= attr}
	end
	

and annotate_and (ANDState s :'aor andstate) : attr andstate * attr =
	let val kids = map annotate_or (#children s)
	    val attrs= map annotationor kids
	    val attr = { expure = (#entry s=empty_action) andalso 
				  List.all #expure attrs,
			 enpure = (#exit  s=empty_action) andalso 
				  List.all #enpure attrs,
			 deeph  = List.all #deeph attrs }
	in (setchildrenand (ANDState s) kids,attr) end

fun annotate_h (s, stenv) = idx_hierarchy (#1(annotate_and s))

fun annotate_sys (sys :('aor,'atr) system) :(attr,'atr) system = 
	let val h = annotate_h (#hierarchy sys)
	in sethierarchy sys h end

end
