(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/Specialize.sml,v
   Comment: Specializer. Executes RL commands on the model
   Release: scope-0-16_final 

     Copyright (C) 2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   Specialize.sml,v 1.6 2003/07/08 15:45:01 wasowski Exp
*)

structure Specialize : TRANS = struct

open RL Absyn

structure S = Symbol
structure SE = SymbolEnv

exception NotFound = SE.NotFound
exception Option = Option.Option

type set = Set.set

type options = rl

type aor = unit
type atr = unit

	
fun spec_venv (vaprop :vaprop env) venv = 
	let val spec = map #1 (SE.toList vaprop)
	in SE.minus (venv,spec) end


fun spec_vartypes  (vaprop :vaprop env ) vartypes =
	let val spec = SE.toList vaprop 

	    fun rm ((na,(t,v,l)),vartypes) = 
			if not (SE.defined vartypes na) then semErr 
				("Variable " ^ S.name na ^ " is used in \
				 \restriction but not declared in the model.")
			else if #2 (SE.lookup vartypes na) <> SimpleType t 
			then semErr("Type mismatch between restriction and model \
				    \for variable " ^ S.name na)
			else SE.remove1 vartypes na

	in foldl rm vartypes spec end
				

fun spec_cenv (vaprop :vaprop env) vartypes cenv = 
	let val specv  = SE.toList vaprop
	    fun is_ext_val (n,(t,SOME _,_)) = 
				(not (#1 (SE.lookup vartypes n))
				handle SE.NotFound => false)
	      | is_ext_val  _ = false
	    val specve = List.filter is_ext_val specv 
	    val cenv1  = List.map (fn (n,(t,v,l))=>(n,value(valOf v))) specve 
			 handle Option => intErr Option 200303174
	in SE.bind cenv cenv1 end

fun spec_exp   (frestr as {fuprop,...} :restr) exp = 
	case exp of
		  Cst c => Cst c
		| Access a => 
			spec_access frestr a
		| Prim1(ope,e) => 
			Prim1(ope, spec_exp frestr e)
		| Prim2(ope,e1,e2) =>
			Prim2(ope, spec_exp frestr e1, spec_exp frestr e2)
		| After(e1,e2) =>
			After(spec_exp frestr e1,spec_exp frestr e2)
		| Call(ou,es) =>
			let val ses = map (spec_exp frestr) es
			    val v = #2 (SE.lookup fuprop ou)
				     handle SE.NotFound => NONE
			in case v of
				  NONE => Call(ou,ses)
				| SOME v => After(exp_seq ses, Cst(value v))
			end 

and spec_access (frestr as {vaprop,...}:restr) acc :exp = 
	case acc of
		  AccessVar v =>
			if SE.defined vaprop v then
				Cst (value(valOf(#2 (SE.lookup vaprop v))))
				handle Option => intErr Option 200303173 
						 (* impossible *)
			else Access acc
		| AccessArr(v,e) =>
			if SE.defined vaprop v then
				semErr "Specialization for arrays not \
					\implemented!"
			else Access(AccessArr(v,spec_exp frestr e))
	

fun spec_command (frestr as {vaprop,...} :restr) cmd =
	case cmd of
		  Expr e => 
			Expr (spec_exp frestr e)	
		| Assgn(AccessVar v,e) => 
			if SE.defined vaprop v 
			then Expr (spec_exp frestr e)
			else Assgn(AccessVar v, spec_exp frestr e)
		| Assgn(AccessArr(v,e1),e2) => 
			Assgn( AccessArr(v,spec_exp frestr e1), 
			       spec_exp frestr e2)

fun spec_action (frestr :restr) (cmd,sexp) :action = 
	let val cmd1 = map (spec_command frestr) cmd
	    val _ = assert (fn()=> List.all (fn (_,ps)=> null ps) sexp)
			"Signal parameters are not implemented!"
	    fun is_t (e,_) = not (Set.member (#nevere frestr,e))
	in (cmd1,List.filter is_t sexp) end

fun spec_tran (frestr :restr) tran :atr tran =
	{ 
		trigger   = #trigger tran,
		scope     = #scope tran,
		pos       = #pos tran,
		neg       = #neg tran,
		guard     = map (spec_exp frestr) (#guard tran),
		action    = spec_action frestr (#action tran),
		targets   = #targets tran,
		annotation= #annotation tran
	}


fun spec_trans (frestr :restr) trans :atr tran list = 
	let fun is_t t = case #trigger t of 
				 Event(e1,_) => not (Set.exists (fn e=> e=e1) 
						 (#nevere frestr))
			       | _ => true
	    val trans1 = List.filter is_t trans
	in map (spec_tran frestr) trans1 end 

fun spec_or  (frestr :restr) (ORState s  :aor orstate) = ORState s
fun spec_and (frestr :restr) (ANDState s :aor andstate) = 
	ANDState {
		name= #name s,
		parent= #parent s,
		children= #children s,
		entry= spec_action frestr (#entry s),
		exit=spec_action frestr (#exit s)
	}


fun spec_hierarchy (frestr :restr) h :aor hierarchy = 
	idx_hierarchy (hmapand (spec_or frestr) (spec_and frestr) (gettree h))

fun fuse_vaprop_venv (vaprop :vaprop env) (venv :venv) :vaprop env =	
	let fun add (v,(t,NONE,l)) = 
			(t,SOME (SE.lookup venv v), l)
	      | add (v,(t,SOME v1,l)) = 
			(t,SOME v1,l) before (if SE.lookup venv v <>v1 
			then warning ("Initialization value and constraint \
				\value in conflict for variable " ^ S.name v) 
			else ())
	in SE.fullmap add vaprop end
	handle NotFound =>  intErr NotFound 200303175

fun spec_otype (fuprop :fuprop env) (ou,(ret,ps,pu)) =
	let val _ = assert (fn()=>not pu) "Purity assumption before specialize?"
	    val pure = #3 (SE.lookup fuprop ou) handle SE.NotFound => false
	in (ret,ps,pure) end
	handle e => intErr e 200303200
	
	

fun spec_system ({vaprop,fuprop,nevere} :restr) (sys :(aor,atr) system) 
							:(aor,atr) system = 
	let val _ = if (not o Set.exists (not o SE.defined (#evtypes sys)))
			nevere then ()
		    else semErr "Event referred to in restriction \
				 \is undefined in the model"
	    val vartypes1 = spec_vartypes vaprop (#vartypes sys)
	    val frestr = { vaprop= fuse_vaprop_venv vaprop (#venv sys),
			   fuprop= fuprop, nevere= nevere }
	in { 
		name     = #name sys,
		fname    = #fname sys,
		vartypes = vartypes1,
		hierarchy= spec_hierarchy frestr (#hierarchy sys),
		evtypes  = SE.minus (#evtypes sys,Set.listItems nevere),
		evgroups = SE.map (fn eg=>Set.difference (eg,nevere)) 
				(#evgroups sys),
		otypes   = SE.fullmap (spec_otype fuprop) (#otypes sys),
		denv     = #denv sys,
		venv     = spec_venv vaprop (#venv sys),
		cenv     = spec_cenv vaprop (#vartypes sys) (#cenv sys),
		trans    = spec_trans frestr (#trans sys) }
		
	end 
		
fun apply1 (rl :rl) (sys :(aor,atr) system, m :(aor,atr) model) = 
	spec_system (SE.lookup (#sys rl) (#name sys)) sys :: m
	handle SE.NotFound => m


fun apply (rl :rl) (mname :string) (m :(aor,atr) model) = 
	foldl (apply1 rl) [] m



end
