(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/VsrAbsyn.sml,v
   Comment: Translation from vsr (textual description oriented) AST to
	    AST defined in Absyn.sml (high level - more visually oriented)
   Release: scope-0-16_final

     Copyright (C) 2001-2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   NOTES:

   1. Variable ranges are lost during translation.

   2. Error reporting is poor (mainly because vsr is not decorated with
      location - probably unavoidable). Errors are reported by impossible or
      asserts. Especially impossible should be used only where it supposed
      to!
      
   3. CRITICAL: do rules are not translated

   4. internal/external properties of expressions are lost

   5. CRITICAL: internal rules are not translated (lost!). AND-State
      containing an internal rules should be translated to a fresh or-state with
      self loop containing the andstate.

   6. CRITICAL: module languge is very poorely treated. All programs are
      passed as systems (should not be so) and no parameters are passed. 
      As a result big real life examples cannot be tested!!!
   
   7. (???) Event groups containing a single event or empty event groups
      currently produce only a warning and no measure is taken against them
      (should be eliminated in fact). We will see how common the warning is.
      Also probably this should be done at Absyn level to avoid excess logic
      in front end (which is not reused among various formats).

      NOTE ON MODULE LANGUAGE: the idea is the vsr files will be grouped in
      two one extra level topstate within one system (so a bunch of calls
      creates a model in Absyn). In this way systems are completely
      independent (as it is in VS) and models may be code generated
      completely separately. Existance of calls is a bit hard to understand
      in VS. I think they could get away without them. 

   7. If efficiency is a bottleneck then rewrite builiding the list of
      transitions in CPS style (half of the work is already done. The only
      thing needed is to add *yet another argument* to mkor and mkand to
      avoid usage of @).
      
   VsrAbsyn.sml,v 1.39 2003/12/15 16:03:29 wasowski Exp
*)

structure VsrAbsyn : CONVIN = struct

  open Absyn List;
  infixr 0 $;

  type source = vsr.file
  type aor    = unit
  type atr    = unit

  structure S = Symbol
  structure SE = SymbolEnv

  type	senv  = S.symbol env
  type stenv  = vsr.stpath env 


  (* strips option from 'a list option. NONE is translated to [] *)

  fun stripOpt NONE	= []
    | stripOpt (SOME l) = l

  fun literal2val (vsr.RealConst s) = ValR (valOf(Real.fromString s))
    | literal2val (vsr.DecConst  i) = ValI i
    | literal2val (vsr.HexConst  i) = ValI i
    | literal2val (vsr.CharConst c) = 
	ValI( ord (String.sub (Option.valOf(String.fromCString c),0)))

  fun autocast2int (ValI i) = i
    | autocast2int (ValR _) = 
	semErr "Real values may not be casted to integers implicitly"
    | autocast2int (ValA a) =
	semErr "Arrays may not be casted to integers implicilty"
		(* FIXME PROBLEM: actually they can in C *)

  (* state names *)

  fun qname (stname : S.symbol list) : string = 
	let fun mk  []	   res = String.concat res
	      | mk (s::st) []  = mk st [S.name s]
	      | mk (s::st) res = mk st (S.name s::"."::res)
	in mk stname [] end

  (* events environment *)

  fun mkevtypes (es : vsr.eventSpec list) (ss : vsr.eventSpec list) = 
	let 
		fun mk (signal :bool) ((n,ps),_) :event * evtype =
			(n, (signal, map #1 (stripOpt ps)))
		val es' = foldl (fn (e,env) => SE.bind1 env (mk false e)) 
				empty es
		val es'' = foldl (fn (e,env) => SE.bind1 env (mk true e)) 
				es' ss 
	in es'' end

  (* event groups environment *)
  
  fun mkevgroups evtypes (gs : vsr.groupSpec list) = 
	let fun mk (name,ps,_) : symbol * evgroup = 
		if SymbolEnv.defined evtypes name then
			semErr ("Event group and event " ^ S.name name ^
				" have the same identifier.") 
		else let val events = stripOpt ps
			 val _ = case events of
					  []  => warning $ "Empty event group:"
							^ S.name name
					| [e] => warning $ "Event group " ^ 
							S.name name ^ 
							" contains only a \
							\single memeber"
					| _ => ()
		in (name,Set.fromList events) end
	in foldl (fn (g,env) => SE.bind1 env (mk g)) empty gs end

  (* output environment *)

  fun mkotypes (lst : vsr.actionSpec list) : otype env	=
	let
		fun mk (f:vsr.funSpec) : output * otype =
			(#2 f, (#1 f,map #1 (stripOpt (#3 f)),false))  
		fun mko (vsr.ActionSpec     (f,_)) = mk f
		  | mko (vsr.ActionTimerSpec(f,_)) = mk f
	in foldl (fn (a,env) => SE.bind1 env (mko a)) empty lst end

  (* variable environment *)	

  fun bindvars extv intv : vartype env * venv = 
	let
		fun rep elem count = tabulate (count,(fn _ =>elem))
		fun mk internal (vsr.VarSimple(typ,id,_,init,_)) 
						: var * vartype * value = 
			let val vtyp = (internal, SimpleType typ)
			    val ini  = case init of
					  SOME l => literal2val l
					| NONE	 => defvalty typ
			in (id,vtyp,ini) end
		  | mk internal (vsr.VarArray(typ,id,n,_,init,_)) =
			let val size = autocast2int(literal2val n) 
			    val ini  = case init of
					  SOME l => map literal2val l
					| NONE	 => rep (defvalty typ) size
			    val vtyp = (internal,ArrayType (typ,size))
			in (id, vtyp, ValA ini) end
		 fun bindvar internal (vsrvar,(vartypes,venv)) =
			let val	(id,typ,ini) = mk internal vsrvar
			    val vartypes1 = SE.bind1 vartypes (id,typ)
			    val venv1	  = SE.bind1 venv	   (id,ini)
			in (vartypes1,venv1) end
		val ex = foldl (bindvar false) (empty,empty) extv
	in foldl (bindvar true) ex intv end



  (* processing expressions *)	

  fun mkcnst (vsr.RealConst s) = CstR( valOf(Real.fromString s))
    | mkcnst (vsr.DecConst  i) = CstI i
    | mkcnst (vsr.HexConst  i) = CstI i
    | mkcnst (vsr.CharConst c) = 
	CstC (String.sub(valOf(String.fromCString c),0))

  fun mkexpr (vsr.ExprLiteral l  ) : exp = Cst	  (mkcnst   l)
    | mkexpr (vsr.ExprFunCall f  )	 = Call   (mkoexp   f)
    | mkexpr (vsr.ExprVar     v  )	 = Access (mkaccess v)
    | mkexpr (vsr.ExprPar     e  )	 = mkexpr e
    | mkexpr (vsr.ExprUnary (ope,e))	 = Prim1  (ope,mkexpr e)
    | mkexpr (vsr.ExprBinary(e1,ope,e2)) = Prim2  (ope,mkexpr e1,mkexpr e2)
  
  and mkoexp (vsr.FunCall (f,pars)) = (f, map mkexpr pars)

  and mkaccess (vsr.SimpleVar v) : access = AccessVar v
    | mkaccess (vsr.SubscriptVar (v,e))   = AccessArr (v,mkexpr e)

  (* processing actions *)

  fun tgtact (vsr.ActionNextState _) = true
    | tgtact _			     = false
  fun outact (vsr.ActionFunCall   _) = true
    | outact _			     = false
  fun sigact (vsr.ActionSignal	  _) = true
    | sigact _			     = false
  fun asgnact(vsr.ActionAssign	  _) = true
    | asgnact _			     = false

  fun mksig (vsr.ActionSignal (e, SOME pars)) = (e, map mkexpr pars)
    | mksig (vsr.ActionSignal (e, NONE)) = (e,[])
    | mksig _ = 
	impossible "Attempt to translate nonsignal vsr action to sexp."

  fun mkout (vsr.ActionFunCall f) = mkoexp f
    | mkout _ =
	impossible "Attempt to translate non-funcall vsr action to oexp."

  fun mkassgn (vsr.ActionAssign (lval,rval,_)) = (mkaccess lval,mkexpr rval)
    | mkassgn _ =
	impossible "Attempt to translate non-assign vsr action to assgn."
    
  fun mkaction (ontran : bool) (rules : vsr.action list) : action = 
	let
		val _ = assert (fn () => 
					ontran orelse 
					not(exists tgtact rules))
				("Force target action is illegal if not a \
					\transition component.")

		val rules   = filter (not o tgtact) rules
		val assgns  = map (Assgn o mkassgn) (filter asgnact rules)
		val outputs = map (Expr o Call o mkout) (filter outact  rules)
		val signals = map  mksig (filter sigact  rules)
	in ( assgns @ outputs, signals ) end

  (* hierarchy: expanding parameters used in given object *)

  (* An ugly hack: I append empty string at the begining of the list to say
  that it has beed already expanded. VERY INEFFICIENT. EACH TRANSITION IS 
  EXPANDED REDUNDANTLY (NO EXPANSION REALLY) FOR TRANSITION'S DEPTH TIME.
  YUCK! FIXME STILL I SEE AN EASY WAY TO AVOID THAT. ALTERNATIVE IS TO MARK
  EACH STATE AS IT IS EXPANDED COMPLETELY, THEN MARK TRANSITIONS AND SPLIT
  THEM INTO POOLS ETC. ACTUALLY THIS WOULD RESULT IN NATURAL DIVISION OF
  TRANSITIONS BY SCOPE. *)

  fun expandst stenv parent [s] = 
	(SE.lookup stenv s handle NotFound => S.symbol ""::s::parent) 
    | expandst stenv parent []	= 
	impossible ("Empty state name for parent " ^ qname parent)
    | expandst stenv parent ss	= 
	if hd ss = S.symbol "" then ss else S.symbol "" ::revAppend(ss, parent)

  fun expandguard stenv parent (vsr.GuardPos  st) = 
			vsr.GuardPos (expandst stenv parent st)
    | expandguard stenv parent (vsr.GuardNeg  st) = 
			vsr.GuardNeg (expandst stenv parent st)
    | expandguard stenv parent g = g

  fun expandact stenv parent (vsr.ActionNextState st) = 
			vsr.ActionNextState(expandst stenv parent st)
    | expandact stenv parent a = a


  fun expandtran (stenv :stenv) (parent :vsr.stpath)
		( {event,condition,actions,tag} :vsr.rule ) : vsr.rule =
	{ event     = event, 
	  condition = map (expandguard stenv parent) condition, 
	  actions   = map (expandact   stenv parent) actions,
	  tag	    = NONE }

  (* full expansion - second pass over transitions to insert 
     or-state names *)

  
  (* this a loop which iterates over 'a list evaluating f until it succeeds.
     'b and 'c are types of additional parameters to f.  The f function used
     here is findor and findand defined below.	The list we iterate over is
     the list of siblings under given style and f performs name matching. *)
  
  fun loop (f : 'a -> 'b -> 'c -> 'd) (b :'b) (c :'c) ([] :'a list) :'d = 
		raise StateNotFound
    | loop f b c (a::atl) = f a b c handle StateNotFound => loop f b c atl

  fun findor (ORState orst) (parent :string) (st :vsr.stpath) = 
	let val this = S.name (#name orst)
	in case st of 
		  s::[] => if (parent ^ "." ^ S.name s = this) then #name orst
			   else loop findand this st (#children orst)
		| s::t	=> if (parent ^ "." ^ S.name s = this) then 
				loop findand this t (#children orst)
			   else loop findand this st (#children orst)
		| [] => impossible "function should not reach this case."
	end

  and findand (ANDState andst) ("" :string) (st :vsr.stpath) :stateid =
	let val this = S.name (#name andst)
	in if S.name (hd st) = this then
		loop findor this (tl st) (#children andst)
	   else impossible "root name unmatched"
	end
    | findand (ANDState andst) parent st  =
	let val this = S.name(#name andst)
	in case st of
		  s::[] => if (parent ^ "." ^ S.name s = this) then #name andst
			   else raise StateNotFound
		| s::t =>  if (parent ^ "." ^ S.name s = this) then
				loop findor this t (#children andst)
			   else raise StateNotFound
		| [] => impossible "function should not reach this case."
	end

  (* ffname - find full name of semi properly qualified state 
	      (a state with some or-states omitted in its name). This
	      function finds a correct fully qualified name. The real job is
	      done by two mutually recursive searchers find_and and find_or. *)

  fun ffname (h :aor hierarchy) ([] :vsr.stpath) : stateid = 
	impossible "A name with no expansion mark and no system (empty name)!"
    | ffname h (pref::st) = 
          (if S.name pref = "" then findand (gettree h) "" (rev st)
 	   else impossible ("A name has not been lifted: " ^ qname st))
	handle Empty => impossible ("There should be at least a root node in\
				    \ qualified state name:" ^ qname st)
	     | StateNotFound => impossible ("Attempt to expand a name not in\
					    \ the hierarchy: " ^ qname st )
	     | e => intErr e 200312141


  (* processing transitions *)	

  fun mktrigger evtypes evgroups (event, SOME (p::ps)) = 
	if SymbolEnv.defined evtypes event then Event (event, p::ps)
	else semErr ("Undeclared event " ^ S.name event)
    | mktrigger evtypes evgroups (name, _ ) = 
	if SymbolEnv.defined evtypes name then Event (name,[])
	else if SymbolEnv.defined evgroups name then EventGroup name
	else semErr ("Undeclared event " ^ S.name name)
		
  fun istarget (vsr.ActionNextState _) = true
    | istarget _		       = false

  fun gettarget (vsr.ActionNextState st) = st
    | gettarget _ = 
	impossible "Should never ask for target of non target actions."

  fun ispos (vsr.GuardPos st) = true 
    | ispos _		      = false
    
  fun isneg (vsr.GuardNeg st) = true
    | isneg _		      = false

  fun getcond (vsr.GuardPos st) = st
    | getcond (vsr.GuardNeg st) = st
    | getcond _ =
	impossible "Should never ask for state value of plain condition."

  fun isexpr (vsr.GuardExpr e ) = true
    | isexpr _	= false

  fun getexpr (vsr.GuardExpr (e,_)) = e
    | getexpr _ = impossible "Should never ask for expression component of \
			     \non-expression guard."

  fun mktran evtypes evgroups (h :aor hierarchy)
		( {event,condition,actions,tag} :vsr.rule ) : atr tran =
	let 
	    fun deor (InclAND s) :aor andstate = s
	      | deor (InclOR  s) = deor $ getstate h $ parentor s

	    val pos1  = filter ispos condition
	    val pos2  = map (ffname h o getcond) pos1
	    
	    val neg1  = filter isneg condition
	    val neg2  = map (ffname h o getcond) neg1
	    
	    val tgts1 = filter istarget actions
	    val tgts2 = map (ffname h o gettarget) tgts1

	    val expr1 = filter isexpr condition

	    val p = hd pos2 handle Subscript => 
			impossible "There should be at least one source!"
	    val t = hd tgts2 handle Subscript => 
			impossible "There should be at least one target!"
	    val l = LCA h [parenth h p, parenth h t]

	in { trigger   = mktrigger evtypes evgroups event,
	     scope     = l,
	     pos       = pos2,
	     neg       = neg2,
	     guard     = map (mkexpr o getexpr) expr1,
	     action    = mkaction true actions,
	     targets   = tgts2,
	     annotation= () } 
	end handle e => intErr e 200312159

  (* processing hierarchy *)

  fun dropuntil (machname :S.symbol) ([] :vsr.stateMachineDef list) = 
	semErr ("State " ^ S.name machname ^ " is undefined!")
    | dropuntil machname (m::ms) = if #name m = machname then m::ms
				   else dropuntil machname ms

  fun mkbasic (stname : vsr.stpath) (st :vsr.state) :aor andstate = 
	ANDState { name = S.symbol (qname stname), 
		   parent = SOME $ S.symbol $ qname $ tl stname, 
		   children = [],
		   entry = empty_action, exit = empty_action }

(* pref is a qualified state name in reversed order: root is last.
   It is used as a prefix for current state name. *)		   

  fun mkor (pref :S.symbol list) (reg :vsr.region option) 
	([] :vsr.stateMachineDef list) (deeph :bool) (stenv :stenv) 
		(denv :denv) :aor orstate * denv * vsr.rule list =

	impossible ("Non existing OR-state of parent " ^ (qname pref))
	
    | mkor pref reg (m::ms) deeph stenv denv	=
	let 
		val (orname,states) = case reg of
			  NONE	=> (S.fresh()::pref, #states m)
			| SOME (s,SOME rs,_) => 
				let fun inreg st= 
					exists (fn s=>s= #name st) rs
				in (s::pref, filter inreg $ #states $ m) end
			| SOME (s,_,_) => impossible ("Region is empty.")

		fun ckDeepHis st = (#stateType st = vsr.DeepHistoryState)
		val hasDeepHis   = exists ckDeepHis states
		val deeph1       = deeph orelse hasDeepHis

		fun mk (st,(kids,denv,tran)) 
			:aor andstate list * denv * vsr.rule list = 
			let 
				val andname = #name st::orname 
			in
			    case #refersTo st of
				  NONE => 
					(mkbasic andname st::kids, denv, tran) 
				| SOME (m,NONE) => 
					let val (kid,denv1,tran1) = mkand
						andname (dropuntil m ms)
						deeph1 empty denv
					in (kid::kids, denv1, tran1 @ tran) end
				| SOME (m,SOME ps) => 
					let val ms1 = dropuntil m ms
					    val mps = #param (hd ms1)
					    val stenv1 = SE.bindZip empty (mps,ps)
					    val (kid,denv1,tran1) = mkand 
							andname ms1 deeph1 stenv1 denv
					in (kid::kids, denv1, tran1 @ tran) end
			end 

		val (kids,denv1,tran1) = foldr mk ([],denv,[]) states

		fun ckInitial st = (#stateType st = vsr.InitialState)
		fun ckHistory st = (#stateType st = vsr.HistoryState)
		val hasInitial	 = exists ckInitial states
		val hasHistory	 = exists ckHistory states 
		val qorname	 = qname orname
		val name	 = S.symbol qorname
		
		val mrkrs = (if hasInitial then 1 else 0) +
			    (if hasHistory then 1 else 0) +
			    (if hasDeepHis then 1 else 0)

		val _ = if mrkrs < 1 then 
				semErr ("State " ^ qorname ^ " is declared to be\
					\ neither history nor initial.")
			else if mrkrs > 1 then
				warning ("State " ^ qorname ^ " contains\
					 \ conflicting markers.")
			else ()

		val zp	= ListPair.zip (states,kids)

		val _   = assert (fn () => not (null states))	
			  ("Empty regions are not allowed: " ^ qorname ^ "!")

		val def = if mrkrs > 0 
			then find (fn (s,_)=>ckInitial s orelse 
					     ckHistory s orelse 
					     ckDeepHis s) zp
			else SOME (hd zp)

		val denv2 = SE.bind1 denv1 (name, nameand (#2 (valOf def)))
	in 
		( ORState { name      = name,
			    parent    = S.symbol $ qname $ tl orname,
			    children  = kids,				  
			    marking   = if hasHistory orelse hasDeepHis orelse deeph 
					then History else Initial,
			    annotation= ()
			  },
		  denv2,tran1 )
	end handle e => intErr e 200312158

(* stname is a qualified state namein reversed order (root is last).
   It is used as a name for state being created. *)	

  and mkand (stname :vsr.stpath) ([] :vsr.stateMachineDef list) (deeph :bool) 
	(stenv :stenv) (denv :denv) 
		:aor andstate * denv * vsr.rule list =

	semErr ("Non existing specification for AND-state " ^ qname stname)

    | mkand stname (m::ms) deeph stenv denv = 
	let 
		fun nonempty (_,SOME [],_) = false
		  | nonempty (_,NONE,_)    = false
		  | nonempty _		   = true

		fun mkreg (reg,(kids,denv,tran)) =
			let val (orst,denv1,tran1) = 
				mkor stname reg (m::ms) deeph stenv denv
			in (orst::kids, denv1, tran1 @ tran) end

		val (regs,empt) = List.partition nonempty (#regions m)
		val qstname	= qname stname
		val _ = if null empt then () 
			else warning $ "Empty regions in state " ^ qstname

		val (kids,denv1,tran1) = case regs of
				  [] => if null $ #states m
					then ([],denv,[]) (*BASIC*)
					else mkreg (NONE,([],denv,[]))
				|  _ => foldr mkreg ([],denv,[]) $ map SOME regs 
				
		val tran2	    = #trsRules m @ tran1
		val tran3	    = map (expandtran stenv stname) tran2
		val parname = case tl stname of 
				  [] => NONE 
				| t  => SOME (S.symbol $ qname $ tl stname)
				
	in 
		( ANDState { name     = S.symbol qstname,
			     parent   = parname,
			     children = kids, 
			     entry    = mkaction false $ concat $ map #1 
						       $ #entryRules m,  
			     exit     = mkaction false $ concat $ map #1
						       $ #exitRules m },
		  denv1, tran3 )
	end handle e => intErr e 200312157

(* creates renaming environment for model constants *)

fun bindconsts (specs :vsr.constSpec list) :const env =
	let val csts = map (fn (sym,lit,tag) => (sym,mkcnst lit)) specs
	in SE.bind empty csts end handle e => intErr e 200312156

(* here it is where hierarchy is built *)	

  fun mkhierarchy (vsr.TopStateMachine(top,_)) 
		  (ms :vsr.stateMachine list) 
			:aor andstate * denv * vsr.rule list =
	let fun stripm (vsr.StateMachine    (d,t)) = d
	      | stripm _ = impossible "Topstate is not unique"
	    val ms1 = top :: map stripm (rev ms)
	in 
		mkand [#name top] ms1 false empty empty 
		handle e => intErr e 200312155
	end
    | mkhierarchy _ _ = impossible "Topstate is a plain state machine."

    fun mksys (sysname :symbol) ({filepath, directives, interface=i, 
		       machines, top} : vsr.program ) : (aor,atr) system = 
	let 
		val (vartypes,venv) = bindvars	(#extvars i) (#intvars i)
		val evtypes  = mkevtypes  (#events  i) (#signals i)
		val evgroups = mkevgroups evtypes (#groups  i)
		val (root,denv,vsrtrans) = mkhierarchy top machines
		val h = idx_hierarchy root
		val trans = map (mktran evtypes evgroups h) vsrtrans
	in 
		{ name	    = sysname,
		  fname     = filepath,
		  hierarchy = h, 
		  vartypes  = vartypes,
		  evtypes   = evtypes,
		  evgroups  = evgroups,  
		  otypes    = mkotypes (#actions i),
		  trans     = prioritize root trans,
		  denv	    = denv,
		  venv	    = venv,
		  cenv      = bindconsts (#constants i) } 
	end handle e => intErr e 200312154


  (* main translating functions *)

  fun translate ( vsr.Project p ) : (aor,atr) model = 
	let val progs = map #3	(#programs p)
	    val names = map #2  (#sysnames p)
	in 
		ListPair.map (fn (p,n) => mksys n p) (progs,names) 
		handle e => intErr e 200312152
	end
    | translate ( vsr.Program p ) : (aor,atr) model = 
	[mksys (Symbol.symbol "system") p]
	handle e => intErr e 200312153
    
end
